"""
Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License.
A copy of the License is located at

http://aws.amazon.com/apache2.0

or in the "license" file accompanying this file. This file is distributed
on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
express or implied. See the License for the specific language governing
permissions and limitations under the License.

"""
import os
import logging
import json
import boto3
from botocore.exceptions import ClientError
from events_iterator import EventsIterator
import numpy as np
from amazon_aed import AmazonAED
from decimal import Decimal


DEBUG_MODE = os.getenv('DEBUG_MODE', 'true')

LOGGER = logging.getLogger(__name__)
os.putenv('LOGGER_NAME', __name__)

if DEBUG_MODE == 'true':
    LOGGER.setLevel(logging.DEBUG)
else:
    LOGGER.setLevel(logging.INFO)

ssmclient = boto3.client('ssm')    
LAST_CHANGE_POINT_SSM_PARAM = str(os.getenv('AED_LAST_CHANGE_POINT',''))

AED_STARTUP = int(os.getenv('AED_STARTUP', '5'))    
AED_ALPHA_LEVEL = float(os.getenv('AED_ALPHA_LEVEL', '0.01'))
AED_CHANGE_DIFFERENCE_WINDOW = int(os.getenv('AED_CHANGE_DIFFERENCE_WINDOW', '5'))
AED_PRACTICAL_SIGNIFICANCE_THRESHOLD = float(os.getenv('AED_PRACTICAL_SIGNIFICANCE_THRESHOLD', '0'))
AED_NOTIFICATION = str(os.getenv('AED_NOTIFICATION',''))
AED_CHANGE_POINTS_TABLE = str(os.getenv('AED_CHANGE_POINTS_TABLE',''))
AED_EVENTS_TABLE = str(os.getenv('AED_EVENTS_TABLE',''))

dynamodb = boto3.resource('dynamodb')
change_points_table = dynamodb.Table(AED_CHANGE_POINTS_TABLE) 

cloud_watch = boto3.client('cloudwatch')
sns = boto3.client('sns')


def lambda_handler(event, context):

    try:

        LOGGER.info(f'Received event {json.dumps(event)}')
        # Only looking at 1 record: batchsize=1
        new_item = event["Records"][0]["dynamodb"]["NewImage"]
        date = new_item["date"]["S"]
        timestamp = new_item["timestamp"]["S"]

        LOGGER.debug(f'date = {date}, timestamp = {timestamp}')

        try:
            # Get the last change point
            last_change_point = ssmclient.get_parameter(Name=LAST_CHANGE_POINT_SSM_PARAM)['Parameter']['Value']
            change_point_date, change_point_timestamp = tuple(last_change_point.split("|"))
            LOGGER.debug(f'last change point date: {change_point_date}, change_point_timestamp: {change_point_timestamp} ')

            # Retrieve from Dynamo, the list of data items since the last change point
            events_iterator = EventsIterator(
                start_date=change_point_date,
                end_date=date,
                timestamp=change_point_timestamp,
                ddb_table_name=AED_EVENTS_TABLE
            )

            # Turn the retrieved entries into a numpy array; and also into a list 
            X = []
            ddb_timeseries = []
            for item in events_iterator:
                X.append(np.float(item['data_point']))
                ddb_timeseries.append((item))

            X = np.array(X)

            aed = AmazonAED()

            aed.streamingMPP(X=X, 
                            startup=AED_STARTUP, 
                            alpha_level=AED_ALPHA_LEVEL, 
                            change_difference_window=AED_CHANGE_DIFFERENCE_WINDOW,
                            confidence_intervals=None, 
                            practical_significance_threshold=AED_PRACTICAL_SIGNIFICANCE_THRESHOLD)

            for i in range(0, len(aed.change_points)):

                change_point = ddb_timeseries[aed.change_points[i]]
                magnitude = aed.magnitudes[i]
                p_value_change_point = aed.p_value_change_points[i]

                LOGGER.info(f'Change point: {change_point["data_point"]}, magnitude = {magnitude}, p_value_change_point = {p_value_change_point}')

                change_points_table.put_item(
                    Item={
                        'date' : change_point['date'],
                        'timestamp': change_point['timestamp'],
                        'data_point': change_point['data_point'],
                        'magnitude': str(magnitude),
                        'p_value': str(p_value_change_point)
                    }
                )

                '''
                # put change data point in CW
                cloud_watch.put_metric_data(
                    MetricData=[
                        {
                            'MetricName': 'TimeSeries_ChangePoint',
                            'Unit': 'None',
                            'Value': float(change_point['data_point']),
                            'StorageResolution': 1
                        },
                    ],
                    Namespace='AED'
                )
                '''
                
                # Send message to SNS topic
                if AED_NOTIFICATION != '':
                    message = {
                            'date' : change_point['date'],
                            'timestamp': change_point['timestamp'],
                            'data_point': change_point['data_point'],
                            'magnitude': str(magnitude),
                            'p_value': str(p_value_change_point)
                        }
                    sns.publish(
                        TopicArn=AED_NOTIFICATION,   # TO DO or is it TargetARN?
                        Message=json.dumps({'default': json.dumps(message)}),
                        Subject='Change point detected',
                        MessageStructure='json'
                    )

                # update parameter
                new_param_value = ddb_timeseries[aed.change_points[0] + 1]

                LOGGER.debug(f'New parameter : {new_param_value}')
                ssmclient.put_parameter(
                    Name=LAST_CHANGE_POINT_SSM_PARAM,
                    Value=f'{new_param_value["date"]}|{new_param_value["timestamp"]}', Type='String',
                    Overwrite=True
                )



        except ClientError as e:
            if e.response['Error']['Code'] == 'ParameterNotFound':
                LOGGER.info('The first data point ever received')
                ssmclient.put_parameter(Name=LAST_CHANGE_POINT_SSM_PARAM, Value=f'{date}|{timestamp}', Type='String')
                '''
                # put change data point in CW
                cloud_watch.put_metric_data(
                    MetricData=[
                        {
                            'MetricName': 'TimeSeries_ChangePoint',
                            'Unit': 'None',
                            'Value': float(change_point['data_point']),
                            'StorageResolution': 1
                        },
                    ],
                    Namespace='AED'
                )
                '''
                # Send message to SNS topic
                if AED_NOTIFICATION != '':
                    sns.publish(
                        TopicArn=AED_NOTIFICATION,   # TO DO or is it TargetARN?
                        Message={
                            'date' : change_point['date'],
                            'timestamp': change_point['timestamp'],
                            'data_point': change_point['data_point'],
                            'magnitude': str(magnitude),
                            'p_value': str(p_value_change_point)
                        },
                        Subject='First change point detected',
                        MessageStructure='json'
                    )
                       

    except:
        LOGGER.exception('Error during the processing of the event')