"""
Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License.
A copy of the License is located at

http://aws.amazon.com/apache2.0

or in the "license" file accompanying this file. This file is distributed
on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
express or implied. See the License for the specific language governing
permissions and limitations under the License.

"""
import boto3
from datetime import (
    datetime,
    timedelta,
)
from boto3.dynamodb.conditions import (
    Key,
)
import itertools
import os
import logging

DEBUG_MODE = os.getenv('DEBUG_MODE', 'true')

logger_name = os.getenv('LOGGER_NAME', __name__)
LOGGER = logging.getLogger(logger_name)


if DEBUG_MODE == 'true':
    LOGGER.setLevel(logging.DEBUG)
else:
    LOGGER.setLevel(logging.INFO)


class EventsIterator:

    """Iterator for looping over events in time series sequence starting form given data point"""
    def __init__(self, start_date, end_date, timestamp, ddb_table_name):
        self.start_date = start_date
        self.end_date = end_date
        self.start_timestamp = timestamp
        self.ddb_table_name = ddb_table_name

        self.ddb_client = boto3.client('dynamodb')
        self.query_paginator = self.ddb_client.get_paginator('query')

        self.__init_keys(start_date, end_date)
        self.__prepare_iterator()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.main_iterator)

    def __init_keys(self, start_date, end_date):
        self.keys = []

        start = datetime.strptime(start_date, '%Y-%m-%d')
        end = datetime.strptime(end_date, '%Y-%m-%d')

        while start < end:
            start = start + timedelta(days=1)

            self.keys.append(start.strftime('%Y-%m-%d'))

    def __prepare_iterator(self):

        first_itarator = self.query_paginator.paginate(
            TableName=self.ddb_table_name,
            Select='ALL_ATTRIBUTES',
            KeyConditionExpression='#date= :d and #timestamp >= :ts',
            ExpressionAttributeValues={':d': {'S': self.start_date}, ':ts': {'S': self.start_timestamp}},
            ExpressionAttributeNames={'#date': 'date', '#timestamp': 'timestamp'}
        )

        first_generator = self.__crete_generator(first_itarator)
        query_iterators = [first_generator]

        for key in self.keys:
            iterator = self.query_paginator.paginate(
                TableName=self.ddb_table_name,
                Select='ALL_ATTRIBUTES',
                KeyConditionExpression='#date= :d',
                ExpressionAttributeValues={':d': {'S': key}},
                ExpressionAttributeNames={'#date': 'date'}
            )
            query_iterators.append(self.__crete_generator(iterator))

        self.main_iterator = itertools.chain(*query_iterators)


    def __crete_generator(self, iterator):
        for page in iterator:
            for item in page["Items"]:
                yield {
                    'date':item['date']['S'],
                    'timestamp':item['timestamp']['S'],
                    'data_point':item['data_point']['N']
                }